/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) Ericsson AB, 2004-2008. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.jvnet.glassfish.comms.clb.core;

import com.sun.enterprise.tools.verifier.tests.app.AppAltDDAppClient;
import java.util.logging.Logger;

import org.jvnet.glassfish.comms.util.LogUtil;

import java.util.Collection;
import java.util.HashMap;
import java.util.logging.Level;


/**
 *
 * @author kshitiz
 */
public class Controller {
    private static final Logger _logger = LogUtil.CLB_LOGGER.getLogger();

    /* map containing name to cluster mapping */
    private HashMap<String, ServerCluster> clusterMap;

    /* map containing name to request group mapping for http request */
    private HashMap<String, RequestGroup> httpRequestGroupMap;

    /* the load-balancing policy */
    private String lbpolicy;

    /* name of Data Centric Rules file */
    private String dcrFileName;

    /* flag indicating which component(SIP/HTTP) of CLB */
    private int lbType;

    /* map containing instance to base 64 encoded key mapping */
    private GlobalInstanceMap globalInstanceMap;

    /* default http request group to handle http request */
    private RequestGroup defaultHttpRequestGroup;

    /* sip request group to handle sip requests */
    private RequestGroup sipRequestGroup;

    /* number of request being currently serviced */
    private long requestCount;

    /* lock to be acquired before modifying the request count */
    private Object requestCountLock;

    /* instance of dcr file update event notifier */
    private DCRFileUpdateEventNotifier dcrFileUpdateEventNotifier;
    private String controllerType;

    /** Creates a new instance of ControllerImpl */
    public Controller() {
        clusterMap = new HashMap<String, ServerCluster>();
        httpRequestGroupMap = new HashMap<String, RequestGroup>();
        requestCountLock = new Object();
        globalInstanceMap = new GlobalInstanceMap();
    }

    public static Controller createInstance() {
        return new Controller();
    }

    public boolean addCluster(ServerCluster cluster) {
        clusterMap.put(cluster.getName(), cluster);

        return true;
    }

    public boolean deleteCluster(ServerCluster cluster) {
        return (clusterMap.put(cluster.getName(), cluster) != null);
    }

    public boolean enableCluster(ServerCluster cluster) {
        return cluster.enableCluster();
    }

    public boolean disableCluster(ServerCluster cluster) {
        return cluster.enableCluster();
    }

    public void addCluster(String clusterName, boolean selfLoadbalance)
        throws CLBRuntimeException {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.adding_cluster",
                    new Object[]{controllerType, clusterName, selfLoadbalance});

        if (getCluster(clusterName) != null) {
            String message = _logger.getResourceBundle().
                    getString("clb.cannot_add_duplicate_cluster") + clusterName;
            throw new CLBRuntimeException(message);
        }

        ServerCluster cluster = new ServerCluster(clusterName, selfLoadbalance);
        addCluster(cluster);
        
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.successfully_added_cluster",
                    new Object[]{controllerType, clusterName});
    }

    public void addInstance(String clusterName, String instanceName,
        boolean enabled, int timeoutInMinutes, String[] listeners)
        throws CLBRuntimeException {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.adding_instance_to_cluster",
                    new Object[]{controllerType, instanceName, clusterName});

        ServerCluster cluster = clusterMap.get(clusterName);

        if (cluster == null) {
            String message = _logger.getResourceBundle().
                    getString("clb.cluster_does_not_exist") + clusterName;
            throw new CLBRuntimeException(message);
        }

        cluster.addInstance(instanceName, enabled, timeoutInMinutes, listeners);
        
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.successfully_added_instance_to_cluster",
                    new Object[]{controllerType, instanceName, clusterName});
    }

    public void addWebModule(String clusterName, String contextRoot)
        throws CLBRuntimeException {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.adding.web.module",
                    new Object[]{ controllerType, contextRoot, clusterName});

        if (lbType == CLBConstants.SIP_CLB) {
            String message = _logger.getResourceBundle().
                    getString("clb.no_web_module_for_sip_clb");
            throw new CLBRuntimeException(message);
        }
        
        if (!contextRoot.startsWith("/")) {
            contextRoot = "/" + contextRoot;
        }

        RequestGroup requestGroup = null;
        
        if(contextRoot.equals(CLBConstants.DEFAULT_CONTEXT_ROOT)){
            if(defaultHttpRequestGroup == null){
                if(_logger.isLoggable(Level.FINE))
                    _logger.log(Level.FINE,
                            "clb.adding.default.web.module",
                            new Object[]{controllerType, clusterName});
                defaultHttpRequestGroup = RequestGroupFactory.createRequestGroup(this,
                    contextRoot, lbType);
            }                
            requestGroup = defaultHttpRequestGroup;
        }
        else{
            requestGroup = httpRequestGroupMap.get(contextRoot);
            if (requestGroup == null) {
                requestGroup = RequestGroupFactory.createRequestGroup(this,
                        contextRoot, lbType);
                httpRequestGroupMap.put(contextRoot, requestGroup);
            }
        }

        ServerCluster cluster = clusterMap.get(clusterName);

        if (cluster == null) {
            String message = _logger.getResourceBundle().
                    getString("clb.cluster_does_not_exist") + clusterName;
            throw new CLBRuntimeException(message);
        }

        requestGroup.addCluster(cluster);
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.successfully.added.web.module",
                    new Object[]{controllerType, contextRoot, clusterName});
    }

    public ServerCluster getCluster(String clusterName) {
        return clusterMap.get(clusterName);
    }

    public Collection<ServerCluster> getAllClusters() {
        return clusterMap.values();
    }

    public ServerInstance getInstance(String clusterName, String instanceName) {
        ServerCluster cluster = clusterMap.get(clusterName);

        if (cluster == null) {
            return null;
        }

        return cluster.getInstance(instanceName);
    }

    public RequestGroup getRequestGroup(String appName) {
        if (lbType == CLBConstants.SIP_CLB) {
            return sipRequestGroup;
        }

        return httpRequestGroupMap.get(appName);
    }
    
    public RequestGroup getDefaultHttpRequestGroup() {
        return defaultHttpRequestGroup;
    }

    public void initialize() throws CLBRuntimeException {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.intializing_the_controller",
                    new Object[]{controllerType});

        dcrFileUpdateEventNotifier = new DCRFileUpdateEventNotifier();

        if (lbType == CLBConstants.SIP_CLB) {
            sipRequestGroup = RequestGroupFactory.createRequestGroup(this, "",
                    lbType);
        }

        for (ServerCluster cluster : clusterMap.values()) {
            cluster.initialize(lbType);
            createClusterHealthMonitor(cluster);
            cluster.initializeClusterRouter(lbpolicy, this);

            if (lbType == CLBConstants.SIP_CLB) {
                sipRequestGroup.addCluster(cluster);
            }
        }

        if (lbType == CLBConstants.SIP_CLB) {
            sipRequestGroup.initializeRouter(lbpolicy);
        } else {
            for (RequestGroup reqGroup : httpRequestGroupMap.values())
                reqGroup.initializeRouter(lbpolicy);
            if(defaultHttpRequestGroup != null)
                defaultHttpRequestGroup.initializeRouter(lbpolicy);
        }

        populateGlobalInstanceMap();

        dcrFileUpdateEventNotifier.registerWithAdminEventListener();

        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.intialized_the_controller",
                    new Object[]{controllerType});
    }

    public void setPolicy(String policy) {
        if(dcrFileName == null)
            lbpolicy = policy;
    }

    public void setDCRFileName(String name) {
        if(name != null && name.trim().length() > 0){
            dcrFileName = name.trim();
            lbpolicy = RouterFactory.LBPolicy_CONSISTENT_HASH;
        }
        else
            dcrFileName = null;
    }

    public String getDCRFileName() {
        return dcrFileName;
    }

    public void setLBType(int lbType) {
        this.lbType = lbType;

        if (lbType == CLBConstants.HTTP_CLB) {
            controllerType = "HTTP CLB";
        } else {
            controllerType = "SIP CLB";
        }
    }

    public int getLBType() {
        return lbType;
    }

    public GlobalInstanceMap getGlobalInstanceMap() {
        return globalInstanceMap;
    }

    private void populateGlobalInstanceMap() throws CLBRuntimeException {
        globalInstanceMap = new GlobalInstanceMap();

        for (ServerCluster cluster : clusterMap.values()) {
            for (ServerInstance instance : cluster.getAllInstances()) {
                globalInstanceMap.putServerInstance(instance.getID(), instance);
            }
        }
    }

    /**
     * This method is called at the time of dynamic reconfig. It detects
     * affected server instances and takes care of quiescing them
     *
     */
    public void reconfig(Controller oldController) {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.reconfiguring_the_controller",
                    new Object[]{controllerType});

        Collection<ServerCluster> oldClusters = oldController.getAllClusters();

        for (ServerCluster oldCluster : oldClusters) {
            ServerCluster newCluster = getCluster(oldCluster.getName());

            if (newCluster != null) {
                newCluster.reconfig(oldCluster);
            }
        }

        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.reconfigured_the_controller",
                    new Object[]{controllerType});
    }

    /**
     * This method is obsolete controller to perform the cleanup before being
     * garbage collected. This waits till all request being serviced by this controller
     * are done. Then it performs the cleanup
     */
    public void cleanUp() {
        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.cleaning_up_the_controller",
                    new Object[]{controllerType});

        boolean servingRequest = true;

        while (servingRequest) {
            synchronized (requestCountLock) {
                servingRequest = !(requestCount == 0);

                if (!servingRequest) {
                    break;
                }

                try {
                    requestCountLock.wait();
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }
        }

        for (ServerCluster cluster : clusterMap.values())
            cluster.cleanup();

        if (lbType == CLBConstants.HTTP_CLB) {
            for (RequestGroup reqGroup : httpRequestGroupMap.values())
                reqGroup.cleanup();
            if(defaultHttpRequestGroup != null)
                defaultHttpRequestGroup.cleanup();
        } else {
            sipRequestGroup.cleanup();
        }
        
        ClusterHealthMonitorFactory.removeUnusedClusterHealthMonitors();

        if(_logger.isLoggable(Level.FINE))
            _logger.log(Level.FINE,
                    "clb.cleaned_up_the_controller",
                    new Object[]{controllerType});
    }

    /**
     * Increment the count of request being serviced by this controller
     */
    public void incrementRequestCount() {
        synchronized (requestCountLock) {
            requestCount++;
        }
    }

    /**
     * Decrement the count of request being serviced by this controller
     */
    public void decrementRequestCount() {
        synchronized (requestCountLock) {
            requestCount--;

            if (requestCount == 0) {
                requestCountLock.notify();
            }
        }
    }

    private void createClusterHealthMonitor(ServerCluster cluster)
        throws CLBRuntimeException {
        ClusterHealthMonitor healthMonitor = ClusterHealthMonitorFactory.get(cluster.getName());
        cluster.markInstancesHealthy(healthMonitor.getHealthyInstances());
    }

    /**
     * Returns an instance of DCRFileUpdateEventNotifier associated with this controller.
     *
     * @return instance of dcr file update event notifier
     */
    public DCRFileUpdateEventNotifier getDCRFileUpdateEventNotifier() {
        return dcrFileUpdateEventNotifier;
    }
}
